@@ -1,5 +1,6 @@
 from hmf import MassFunction
-from fitting_functions import Fits
+from fitting_functions import get_fit
+import fitting_functions as fits
 from cosmo import Cosmology
 from transfer import Transfer
 from sample import sample_mf
\ No newline at end of file
