@@ -14,7 +14,6 @@ import scipy.integrate as intg
 import collections
 import cosmolopy as cp
 import itertools
-
 import logging
 
 logger = logging.getLogger('hmf')
@@ -368,79 +367,111 @@ def growth_factor(z, cdict):
     return growth
 
 
+def get_best_param_order(kls, q="dndm", **kwargs):
+    a = kls(**kwargs)
+
+    if isinstance(q, basestring):
+        getattr(a, q)
+    else:
+        for qq in q:
+            getattr(a, qq)
+
+    final_list = []
+    final_num = []
+    for k, v in a._Cache__recalc_par_prop.iteritems():
+        num = len(v)
+        for i, l in enumerate(final_num):
+            if l >= num:
+                break
+        else:
+            final_list += [k]
+            final_num += [num]
+            continue
+        final_list.insert(i, k)
+        final_num.insert(i, num)
+    return final_list
+
 def get_hmf(required_attrs, get_label=True, **kwargs):
     """
     Yield :class:`hmf.MassFunction` objects for all combinations of parameters supplied.
     """
     from hmf import MassFunction
-    import re
 
-    order = ["delta_h", "delta_wrt", "delta_c", "user_fit", "mf_fit", "cut_fit",
-             "z2", "nz", "z", "M", "wdm_mass", "sigma_8", "n", "lnk", "transfer_fit"][::-1]
-
-    ordered_kwargs = collections.OrderedDict([])
-    for item in order:
-        try:
-            if isinstance(kwargs[item], (list, tuple)):
-                ordered_kwargs[item] = kwargs.pop(item)
-        except KeyError:
-            pass
-    # # add the rest in any order
-    for k in kwargs:
-        if isinstance(kwargs[k], (list, tuple)):
-            ordered_kwargs[k] = kwargs.pop(k)
-
-    ordered_list = [ordered_kwargs[k] for k in ordered_kwargs]
-
-    # # Now ordered_kwargs contains an ordered dict of list values, and kwargs
-    # # has an unordered dict of singular values
-
-    final_list = [dict(zip(ordered_kwargs.keys(), v)) for v in itertools.product(*ordered_list)]
-
-    # We want the highest possible wanted attribute
-    if not isinstance(required_attrs, (list, tuple)):
-        attribute = required_attrs
-    else:
-        attribute = "dndm"
-        mattrs = ["ngtm", "mgtm", "mltm", "nltm", "how_big", "dndlnm", "dndlog10m", "dndm",
-                 "fsigma", "n_eff", "lnsigma", "sigma", "_dlnsdlnm", "_sigma_0", "M"]
-        kattrs = ["nonlinear_power", "delta_k", "power", "transfer", "lnk",
-                   "_lnP_0", "_lnP_cdm_0", "_lnT_cdm", "_unnormalised_lnP",
-                   "_unnormalised_lnT"]
-        for a in mattrs + kattrs:
-            if a in required_attrs:
-                attribute = a
-                break
-
-    h = MassFunction(**kwargs)
-
-    for vals in final_list:
-        h.update(**vals)
-        if attribute in mattrs:
-            getattr(h, attribute)
-        elif attribute in kattrs:
-            getattr(h.transfer, attribute)
-        if get_label:
-            if len(final_list) > 1:
-                label = str(vals)
-            elif kwargs:
-                label = str(kwargs)
+    if type(required_attrs) == str:
+        required_attrs = [required_attrs]
+    lists = {}
+    for k, v in kwargs.items():
+        if isinstance(v, (list, tuple)):
+            if len(v) > 1:
+                lists[k] = kwargs.pop(k)
             else:
-                label = h.mf_fit
+                kwargs[k] = v[0]
 
-
-            label = label.replace("{", "").replace("}", "").replace("'", "")
-            label = label.replace("_", "").replace(": ", "").replace(", ", "_")
-            label = label.replace("mffit", "").replace("transferfit", "").replace("delta_wrt", "").replace("\n", "")
-
-            # The following lines transform the M and lnk parts
-            while "[" in label:
-                label = re.sub("[\[].*?[\]]", "", label)
-            label = label.replace("array", "")
-            label = label.replace("M()", "M(" + str(np.log10(h.M[0])) + ", " + str(np.log10(h.M[-1])) + ", " +
-                          str(np.log10(h.M[1]) - np.log10(h.M[0])) + ")")
-            label = label.replace("lnk()", "lnk(" + str(h.transfer.lnk[0]) + ", " + str(h.transfer.lnk[-1]) + ", " +
-                          str(h.transfer.lnk[1] - h.transfer.lnk[0]) + ")")
-            yield h, label
+    x = MassFunction(**kwargs)
+    if not lists:
+        if get_label:
+            yield [[getattr(x, a) for a in required_attrs], x, ""]
         else:
-            yield h
+            yield [[getattr(x, a) for a in required_attrs], x]
+
+    if len(lists) == 1:
+        for k, v in lists.iteritems():
+            for vv in v:
+                x.update(**{k:vv})
+                print "in tools: ", x.mf_fit, k, vv
+                if get_label:
+                    yield [getattr(x, a) for a in required_attrs], x, make_label({k:vv})
+                else:
+                    yield [getattr(x, a) for a in required_attrs], x
+    elif len(lists) > 1:
+        # should be really fast.
+        order = get_best_param_order(MassFunction, required_attrs,
+                                     transfer_fit="BBKS",
+                                     lnk_min=-5,
+                                     lnk_max=5,
+                                     dlnk=1.0,
+                                     Mmin=10,
+                                     Mmax=12,
+                                     dlog10m=0.5)
+
+        ordered_kwargs = collections.OrderedDict([])
+        for item in order:
+            try:
+                if isinstance(lists[item], (list, tuple)):
+                    ordered_kwargs[item] = lists.pop(item)
+            except KeyError:
+                pass
+
+        # # add the rest in any order (there shouldn't actually be any)
+        for k in lists.items():
+            if isinstance(lists[k], (list, tuple)):
+                ordered_kwargs[k] = lists.pop(k)
+
+        ordered_list = [ordered_kwargs[k] for k in ordered_kwargs]
+        final_list = [dict(zip(ordered_kwargs.keys(), v)) for v in itertools.product(*ordered_list)]
+
+        for vals in final_list:
+            x.update(**vals)
+            if not get_label:
+                yield [[getattr(x, q) for q in required_attrs], x]
+
+            else:
+                yield [[getattr(x, q) for q in required_attrs], x, make_label(vals)]
+
+def make_label(d):
+    label = ""
+    for key, val in d.iteritems():
+        if isinstance(val, basestring):
+            label += val + ", "
+        elif isinstance(val, dict):
+            for k, v in val.iteritems():
+                label += "%s: %s, " % (k, v)
+        else:
+            label += "%s: %s, " % (key, val)
+
+    # Some post-formatting to make it look nicer
+    label = label[:-2]
+    label = label.replace("__", "_")
+    label = label.replace("_", ".")
+
+    return label
